# -*- coding: utf-8 -*-

"""
/***************************************************************************
 illini_drainage_tools
                                 A QGIS plugin
 Performs Specific Draiange Related Tasks and Analysis on a Site
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2022-03-15
        copyright            : (C) 2022 by FALASY  Anamelechi
        email                : fvw.services@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

__author__ = 'FALASY  Anamelechi'
__date__ = '2022-03-15'
__copyright__ = '(C) 2022 by FALASY  Anamelechi'

# This will get replaced with a git SHA1 when you do a git archive

__revision__ = '$Format:%H$'

import processing
import os, math
import inspect
import time
import qgis.utils
import numpy as np

from qgis.gui import *
from PyQt5 import QtWidgets
from collections import Counter
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtCore import QCoreApplication, QVariant

from qgis.core import QgsProcessing
from qgis.core import QgsProcessingAlgorithm
from qgis.core import QgsProcessingMultiStepFeedback
from qgis.core import QgsProcessingParameterRasterLayer
from qgis.core import QgsProcessingParameterFolderDestination
from qgis.core import QgsProcessingParameterFileDestination
from qgis.core import QgsProcessingParameterVectorDestination
from qgis.core import QgsProcessingParameterExtent
from qgis.core import QgsProcessingParameterEnum
from qgis.core import QgsProcessingParameterRasterLayer
from qgis.core import QgsProcessingParameterFeatureSource
from qgis.core import QgsProcessingParameterFeatureSink
from qgis.core import QgsProcessingParameterBoolean
from qgis.core import QgsProcessingParameterVectorLayer
from qgis.core import QgsProcessingParameterNumber
from qgis.core import QgsProcessingParameterPoint
from qgis.core import QgsProcessingParameterField
from qgis.core import QgsProcessingParameterCrs
from qgis.core import QgsCoordinateReferenceSystem
from qgis.core import QgsFeatureSink
from qgis.core import QgsFeatureRequest
from qgis.core import QgsVectorLayer
from qgis.core import QgsLineSymbol
from qgis.core import QgsProperty
from qgis.core import QgsProcessingLayerPostProcessorInterface


from qgis.core import (edit,QgsField, QgsFeature, QgsPointXY, QgsWkbTypes, QgsGeometry, QgsFields)

class ElevationAlgorithm(QgsProcessingAlgorithm):            
        
    def tr(self, string):
        return QCoreApplication.translate('Processing', string)
        
    def createInstance(self):
        return ElevationAlgorithm()
        
    def name(self):
        return 'l. Network Elevation Exports'

    def displayName(self):
        return self.tr(self.name())

    def group(self):
        return self.tr(self.groupId())

    def groupId(self):
        return ''
        
    def icon(self):
        cmd_folder = os.path.split(inspect.getfile(inspect.currentframe()))[0]
        icon = QIcon(os.path.join(os.path.join(cmd_folder, 'logo.png')))
        return icon
        
    def shortHelpString(self):
        return self.tr("""This Tool generates elevation points for each line segment of the Tile Network.
        
        Workflow: 
        1. Select a DEM Layer and the Tile Line Layer from "Cumulative Flow Lengths". This is a follow-up from "Routine K"
        2. Select a reference field for generating the elevation points from
        3. Select a desired coordinate reference system for displaying the generated elevation points
        4. Save the output files (optional)
        5. Click on \"Run\"
        
        The script will gives out three outputs.         
                
        The help link in the Graphical User Interface (GUI) provides more information about the plugin.             
        """)    
        
    def helpUrl(self):
        return "https://publish.illinois.edu/illinoisdrainageguide/files/2022/06/PublicAccess.pdf"  
    
    
    def initAlgorithm(self, config=None):        
        self.addParameter(QgsProcessingParameterRasterLayer('MDFT', 'Original Field DEM',  defaultValue=None))        
        self.addParameter(QgsProcessingParameterVectorLayer('LineSegmentLayer', 'Tile Lines: from Cumulative Flow Lengths', types=[QgsProcessing.TypeVectorLine], defaultValue=None))
        self.addParameter(QgsProcessingParameterField('FGHTY', 'Field to Calculate [TILE_ID]', parentLayerParameterName = 'LineSegmentLayer', type = QgsProcessingParameterField.Any,)) 
        self.addParameter(QgsProcessingParameterCrs('CRS', 'Coordinate Reference System', defaultValue='EPSG:3435'))
        self.addParameter(QgsProcessingParameterVectorDestination('FinalFields', 'Final Reference Fields', type=QgsProcessing.TypeVectorAnyGeometry, createByDefault=True, defaultValue=None))
        self.addParameter(QgsProcessingParameterVectorDestination('EndpointElevations', 'End Point Elevations', type=QgsProcessing.TypeVectorAnyGeometry, createByDefault=True, defaultValue=None))
        self.addParameter(QgsProcessingParameterVectorDestination('TerrainProfiles', 'Network Elevation Points', type=QgsProcessing.TypeVectorAnyGeometry, createByDefault=True, defaultValue=None))
        self.addParameter(QgsProcessingParameterFolderDestination('Splitty', 'Network Spreadsheet', createByDefault=True, defaultValue=None))
                     
                        
    def processAlgorithm(self, parameters, context, model_feedback):
        # Use a multistep feedback, so that individual child algorithm progress reports are adjusted for the
        # overall progress through the model    
        feedback = QgsProcessingMultiStepFeedback(9, model_feedback)
        results = {}
        outputs = {}                                       
        
        # Final Retained Fields of Interest
        retained_params = {"INPUT": parameters['LineSegmentLayer'], "FIELDS": ['Tile_ID', 'Tile_TO', 'Elev_First', 'Elev_Last', 'LENGTH', 'FLOW_LENGTH', 'FLOW_LENGT', 'Seg_Slope', 'FLOW_LINE', 'TILE_FLOW', 'FLOW_ORDER', 'Tile_ORDER', 'BURY_ORDER', 'SIZING_ID'], "OUTPUT": parameters['FinalFields']}                   
        
        feedback.setCurrentStep(1)
        if feedback.isCanceled():
            return {}
        
        outputs['RetainFields'] = processing.run('native:retainfields', retained_params, context=context, feedback=feedback, is_child_algorithm=True) #1
        results['RetainFields'] = outputs['RetainFields']['OUTPUT']
        
        # Extract Vertex
        vertex_params = {'INPUT': results['RetainFields'], 'OUTPUT': QgsProcessing.TEMPORARY_OUTPUT}
                        
        feedback.setCurrentStep(2)
        if feedback.isCanceled():
            return {}

        outputs['EndpointVertex'] = processing.run('native:extractvertices', vertex_params, context=context, feedback=feedback, is_child_algorithm=True) #2
        results['EndpointVertex'] = outputs['EndpointVertex']['OUTPUT']
        
        # Add X & Y Fields
        xyadds_params = {'INPUT': results['EndpointVertex'], 'CRS': parameters['CRS'], 'OUTPUT': QgsProcessing.TEMPORARY_OUTPUT}
                        
        feedback.setCurrentStep(3)
        if feedback.isCanceled():
            return {}

        outputs['AddXYfields'] = processing.run('native:addxyfields', xyadds_params, context=context, feedback=feedback, is_child_algorithm=True) #3
        results['AddXYfields'] = outputs['AddXYfields']['OUTPUT']
        
        # Add Field ID
        idz_params = {
            'FIELD_LENGTH': 11,
            'FIELD_NAME': 'Point_ID',
            'FIELD_PRECISION': 3,
            'FIELD_TYPE': 1,
            'FORMULA': '$id',
            'INPUT': results['AddXYfields'],
            'OUTPUT': QgsProcessing.TEMPORARY_OUTPUT
        }
        feedback.setCurrentStep(4)
        if feedback.isCanceled():
            return {}

        outputs['AddFieldID'] = processing.run('native:fieldcalculator', idz_params, context=context, feedback=feedback, is_child_algorithm=True) #4
        results['AddFieldID'] = outputs['AddFieldID']['OUTPUT']        
                
        # Endpoint Fields of Interest
        endpoint_params = {"INPUT": results['AddFieldID'], "FIELDS": ['Point_ID', 'x', 'y'], "OUTPUT": QgsProcessing.TEMPORARY_OUTPUT}
                                
        feedback.setCurrentStep(5)
        if feedback.isCanceled():
            return {}

        outputs['FinalEndPoints'] = processing.run('native:retainfields', endpoint_params, context=context, feedback=feedback, is_child_algorithm=True) #5
        results['FinalEndPoints'] = outputs['FinalEndPoints']['OUTPUT']
        
        # Add Raster Values to Points
        elev_params = {'SHAPES': results['FinalEndPoints'], 'GRIDS': parameters['MDFT'], 'RESAMPLING': 0, 'RESULT': parameters['EndpointElevations']}
                        
        feedback.setCurrentStep(6)
        if feedback.isCanceled():
            return {}

        outputs['ExtractElevationPoints'] = processing.run('saga:addrastervaluestopoints', elev_params, context=context, feedback=feedback, is_child_algorithm=True) #6
        results['ExtractElevationPoints'] = outputs['ExtractElevationPoints']['RESULT']
        
        # Line Terrain Profiles
        line_params = {'DEM': parameters['MDFT'], 'LINES': results['RetainFields'], 'NAME': parameters['FGHTY'], 'PROFILE': QgsProcessing.TEMPORARY_OUTPUT, 'PROFILES': QgsProcessing.TEMPORARY_OUTPUT}
                        
        feedback.setCurrentStep(7)
        if feedback.isCanceled():
            return {}

        outputs['ProfilesFromLines'] = processing.run('saga:profilesfromlines', line_params, context=context, feedback=feedback, is_child_algorithm=True) #7
        results['ProfilesFromLines'] = outputs['ProfilesFromLines']['PROFILE']               
        
        # Rename Field Name        
        rename_params = {'INPUT': results['ProfilesFromLines'], 'FIELD': 'Z', 'NEW_NAME': 'SURF_ELEV', 'OUTPUT': parameters['TerrainProfiles']}
        
        feedback.setCurrentStep(8)
        if feedback.isCanceled():
            return {}
            
        outputs['RenameTableField'] = processing.run('qgis:renametablefield', rename_params, context=context, feedback=feedback, is_child_algorithm=True) #8
        results['RenameTableField'] = outputs['RenameTableField']['OUTPUT']             
               
        # Split and Save as CSV Files
        split_params = {'INPUT': results['RenameTableField'], 'FIELD': 'LINE_ID', "FILE_TYPE": 4, 'OUTPUT': parameters['Splitty']}          
                                              
        feedback.setCurrentStep(9)
        if feedback.isCanceled():
            return {}        
        
        outputs['SplitToCSV'] = processing.run('qgis:splitvectorlayer', split_params, context=context, feedback=feedback, is_child_algorithm=True) #9
        results['SplitToCSV'] = outputs['SplitToCSV']['OUTPUT']  

        if context.willLoadLayerOnCompletion(results['RetainFields']):
            context.layerToLoadOnCompletionDetails(results['RetainFields']).setPostProcessor(LinePostProcessor.create())
        
        return results          
    
class LinePostProcessor(QgsProcessingLayerPostProcessorInterface):

    instance = None

    def postProcessLayer(self, layer, context, feedback):
        if not isinstance(layer, QgsVectorLayer):
            return
        renderer = layer.renderer().clone()
        symbol = QgsLineSymbol.createSimple({'line_color': '0,0,0,255', 'line_width': '0.66', 'line_style': 'solid'})
        renderer.setSymbol(symbol)
        layer.setRenderer(renderer)

    @staticmethod
    def create() -> 'LinePostProcessor':
        LinePostProcessor.instance = LinePostProcessor()
        return LinePostProcessor.instance